//************************************************************************
// T6963.C
// TextRoutinen fr LCD-Displays mit T6963
//
// Hier fr Winstar WG24064A
//
// Ausgelegt bisher nur auf diesen Displaytyp
// Alle Variablen zur Koordinatenbergabe sind 8Bit lang.
// Das reicht fr Displays mit 128x128, 240x64 und 240x128.
//
// Der Nullpunkt des Displays liegt oben,links.
//
// Was witziges nebenbei: Wenn das Display von einer Leuchtstofflampe
// Energiesparlampe beleuchtet wird flimmert es unangenehm !
// Interferenzerscheinung.
// Da hat Winstar eine schlechte Quarzfrequenz gewhlt.
// Hlt man eine Taschenlampe drauf ist es vllig ruhig.
//
// hk@holger-klabunde.de
// http://www.holger-klabunde.de
// Compiler MCC18
// 10.08.2005
//************************************************************************

#include <p18cxxx.h>
#include "T6963.h"

unsigned char xcursor,ycursor;
unsigned char cgram[9];  //cgram[8] ist die CG-RAM Adresse, 0..7 die Daten

//##########################################################
//StatusBits STA0,1 checken
unsigned char CheckBusy(void)
{
 unsigned char by;

 A0=COMMAND;
 DATA_DIR_IN();
 do{
    RD=0;
    by=RD_DATA();
    by&=0x03;
    RD=1;
   }while(by!=3);
   
 DATA_DIR_OUT(); //Alles wieder auf Ausgnge
 return by;
}

//##########################################################
//Datenbyte schreiben
void WriteData(unsigned char dat)
{
 CheckBusy();
 
 A0=DATA;
 WR_DATA(dat); //Daten schreiben
 WR=0;
 WR=1;
}

//##########################################################
//Kommandobyte schreiben
void WriteCommand(unsigned char comand)
{
 CheckBusy();

 A0=COMMAND;
 WR_DATA(comand); //Kommando schreiben
 WR=0;
 WR=1;
}

//###############################################################
//Schreibt ein Datenbyte und ein Kommandobyte
//Bisher ist das Kommandobyte immer 0xC0 fr Write+Autoincrement
//Deshalb keine Parameterbergabe fr 'command' mehr !
//Spart Speicher
void WriteData1(unsigned char dat)
{
 WriteData(dat); //Daten schreiben
 WriteCommand(0xC0); //Kommando schreiben
}

//##########################################################
// Kompletten Bildschirm lschen
// ClearScreen() dauert ca. 5-6ms
void ClearScreen(void)
{
 unsigned char i;

//Write text home address
//Set address pointer = Text home
 LCDPos(1,1);

//ClearText RAM
//Witzigerweise bleibt der Inhalt des Text RAM
//ohne Stromversorgung noch recht lange erhalten ;) 
//Kleiner Trick um fr i nicht uns16 nehmen zu mssen.
//i mte bis 320 zhlen bei 240x64 Pixel Display
//Das hier sieht so aus als wrde nur ein Achtel des Displays
//gelscht werden. Es werden aber 8 Bytes mit increment geschrieben !
 for(i=0; i<(COLS * LCD_HEIGHT/64); i++)
  {
   WriteData1(0x00); WriteData1(0x00);
   WriteData1(0x00); WriteData1(0x00);
   WriteData1(0x00); WriteData1(0x00);
   WriteData1(0x00); WriteData1(0x00);
  }   

//Write text home address
//Set address pointer = Text home
 LCDPos(1,1);
}

//##########################################################
void InitDisplay(void)
{
 unsigned char i;

 DATA_DIR_OUT();
 WR_DATA(0);
 
 WR=1; RD=1; A0=DATA;
 WR_TRIS=0;
 RD_TRIS=0;
 A0_TRIS=0;
 
//Write text home address=0x0000
 WriteData(0);
 WriteData(0);
 WriteCommand(0x40);

//Text area column set= Setze Textspalten 
 WriteData(COLS);
 WriteData(0);
 WriteCommand(0x41);

//Set Offset Register= Setze Startadresse CGRAM
//Offset 0x02 -> 0x1400 Startadresse fr Zeichen ab 0x80 
 WriteData(CGRAM_OFFSET);
 WriteData(0);
 WriteCommand(0x22);

//set display mode = OR
 WriteCommand(0x80);

//Text only display !!
 WriteCommand(0x94);

//8 Sonderzeichen fr SpektrumAnalyzer ins CGRAM schreiben

//0x80 nicht benutzt !
 cgram[0]=0;
 for(i=1; i<9; i++) cgram[i]=0x00;
 MakeCGRam();

//0x81-0x88
 for(i=8; i>0; i--)
  {
   cgram[0]++;
   cgram[i]=BAR;
   MakeCGRam();
  }

 ClearScreen();
}

//##########################################################
// Zeichen von Windows nach T6963 konvertieren
unsigned char ConvertText(unsigned char ch)
{
 unsigned char tmp;

 tmp=ch;

//zwei Abfragen fr die Bargraphen vom Spektrum-Analyzer
//sendet der wirklich bis 8 ? Ohne <= gehts jedenfalls nicht !

//Crystal sendet 128-135 fr die CGRAM Zeichen
// if(tmp<=8) { tmp+=0x80; return tmp; } //CGRAM !
 if(tmp==0xFF) { tmp=0x88; return tmp; } //0xFF heit ausgeflltes Zeichen

//Vom T6963 darstellbare Zeichen
//Das hier vor der bersetzungstabelle knnte noch Probleme geben !
 if(tmp>=0x20 && tmp<=0x7F) { tmp-=0x20; return tmp; }

//Die case-Werte sind bezogen auf den Windows-SystemFont !!!
//Vom LCDPlugin (leider)
//	csText.Replace( '', (char)0xE1 );
//	csText.Replace( '', (char)0xE1 );
//	csText.Replace( '', (char)0xEF );
//	csText.Replace( '', (char)0xEF );
//	csText.Replace( '', (char)0xF5 );
//	csText.Replace( '', (char)0xF5 );
//	csText.Replace( '', (char)0xE2 );

//bersetzungstabelle fr alle anderen Zeichen
//0xFF ist reserviert als Kennzeichen fr Fehler !!
  switch(tmp)
   {
    case 0xE1: tmp=0x6E; break; // 
//     case 196: tmp=0x6E; break; // 
//     case 228: tmp=0x64; break; // 
    case 0xEF: tmp=0x79; break; // 
//     case 214: tmp=0x79; break; // 
//     case 246: tmp=0x74; break; // 
    case 0xF5: tmp=0x7A; break; // 
//     case 220: tmp=0x7A; break; // 
//     case 252: tmp=0x61; break; // 
    case 0xE2: tmp=0x7E; break; //   naja sieht einigermaen so aus
//     case 223: tmp=223; break; // 

    case 128 : tmp=0x80; break; //Umlenkung ins Character RAM
    case 129 : tmp=0x81; break; //Umlenkung ins Character RAM
    case 130 : tmp=0x82; break; //Umlenkung ins Character RAM
    case 131 : tmp=0x83; break; //Umlenkung ins Character RAM
    case 132 : tmp=0x84; break; //Umlenkung ins Character RAM
    case 133 : tmp=0x85; break; //Umlenkung ins Character RAM
    case 134 : tmp=0x86; break; //Umlenkung ins Character RAM
    case 135 : tmp=0x87; break; //Umlenkung ins Character RAM
    default: tmp=0xFF; break; //alle anderen Zeichen ungltig
   }

 return tmp;
}

//##########################################################
// Textcursor positionieren an Position zeile,spalte
void LCDPos(unsigned char zeile, unsigned char spalte)
{
 unsigned char by;
 unsigned int tmp;

 if(spalte>COLS || zeile>ROWS) return; //Position auerhalb des Displays

 by=zeile-1;
 
 tmp=COLS*(unsigned int)by;
 tmp+=(spalte-1);

 by=(unsigned char)(tmp&0xFF);
 WriteData(by);
 by=(unsigned char)(tmp>>8);
 WriteData(by);
 WriteCommand(0x24);

 xcursor=spalte;
 ycursor=zeile;
}

//##########################################################
// Einen Buchstaben schreiben.
void LCDChar(unsigned char ch)
{
 unsigned char ps;

 ps=ch;

//Es scheint besser zu sein 0x0D einfach zu ignorieren
 if(ps==0x0D) { return; } //Wagenrcklauf
// if(ps==0x0A) { return; } //Zeilenvorschub

 if(ps==0x0A) //Nimmt 0x0A als NEWLINE ! d.h. CR+LF
  {
   ycursor++;
   if(ycursor>ROWS) ycursor=1;
   LCDPos(ycursor,1);
   return;
  } //Zeilenvorschub

 ps=ConvertText(ps); //bersetzungstabelle aufrufen
 if(ps!=0xFF)
  {
   //verhindern das auerhalb des Displays geschrieben wird
   if(xcursor<=COLS && ycursor<=ROWS)
    {
     WriteData1(ps); //Write ps
     xcursor++;  //Cursor ein Zeichen weiter
    } 
  
  }//if(ps!=0xFF)

}

//##########################################################
//selbstdefinierte Zeichen ins Character RAM schreiben
//die Zeichen mssen vorher ins Array cgram[] gelegt werden
//cgram[8] ist die Nummer unter der das Zeichen angesprochen
//werden soll minus 0x80. Dann folgen die Bytes von oben
//im Zeichen nach unten
void MakeCGRam(void)
{
 unsigned char i,by;

 by=cgram[0]*8;

// vorher das hier:
// WriteData2(adress,0x24);  //Zeiger auf die Adresse setzen

 WriteData(by); //Daten schreiben

 WriteData(0x14); //Daten schreiben

 WriteCommand(0x24); //Kommando schreiben

 for(i=1; i<9; i++) WriteData1(cgram[i]);

 LCDPos(ycursor,xcursor);  //extrem wichtig ! Textcursor wiederherstellen !
}
